/**
*  drivers/mtd/nand/nand01gw3b2b.c
*
*  Copyright (C) 2007 MSc
*
*  Derived from drivers/mtd/nand/h1910.c
*       Copyright (C) 2002 Marius Grger (mag@sysgo.de)
*       Copyright (c) 2001 Thomas Gleixner (gleixner@autronix.de)
*
* $Id: h1910.c,v 1.6 2005/11/07 11:14:30 gleixner Exp $
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
*  Overview:
*   This is a device driver for the NAND flash device found on the
*   fb4 board which utilizes the ST nand01gw3b2b part. This is
*   a 1Gibit (128MiB x 8 bits) NAND flash device.
*
*  Pin assignment on the pio:
*  0...CLE...Command latch enable, active high
*  1...ALE...Address latch enable, active high
*  2...NCE...Chip enable, active low
*  3...NWP...Write protect, active low
*  4...NU....Not used, always 0
*  5...NU....Not used, always 0
*  6...NU....Not used, always 0
*  7...RB....Ready/Busy, Ready=1,Busy=0
*
*/

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/nios.h>

/*
 * MTD structure for fb4 board
 */
static struct mtd_info *nand01gw3b2b_nand_mtd = NULL;

/*
 * Module stuff
 */

#ifdef CONFIG_MTD_PARTITIONS

/*
 * Define static partitions for flash device
 */
#define NUM_PARTITIONS 3
#define NAND_FLASH_SIZE_MB 128
#define NAND_FLASH_CONFIG_PARTITION_SIZE_MB 2
#define NAND_FLASH_USBDISK_SIZE_MB 10 

static struct mtd_partition partition_info[] = {
	{ name: "nand01gw3b2b NAND Flash Config",
		  offset: 0,
		  size: NAND_FLASH_CONFIG_PARTITION_SIZE_MB * 1024 * 1024
	},
	{ name: "nand01gw3b2b NAND Flash USBDisk",
		  offset: NAND_FLASH_CONFIG_PARTITION_SIZE_MB * 1024 * 1024,
		  size: NAND_FLASH_USBDISK_SIZE_MB * 1024 * 1024
	},
	{ name: "nand01gw3b2b NAND Flash Image",
		  offset: (NAND_FLASH_USBDISK_SIZE_MB + NAND_FLASH_CONFIG_PARTITION_SIZE_MB) * 1024 * 1024,
		  size: (NAND_FLASH_SIZE_MB - NAND_FLASH_USBDISK_SIZE_MB - NAND_FLASH_CONFIG_PARTITION_SIZE_MB) * 1024 * 1024
	}
};

#endif









/**
 * nand_read_byte - [DEFAULT] read one byte from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 8bit buswith
 */
 /*
static u_char nand_read_byte(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	return readb(this->IO_ADDR_R);
}
*/

/**
 * nand_write_byte - [DEFAULT] write one byte to the chip
 * @mtd:	MTD device structure
 * @byte:	pointer to data byte to write
 *
 * Default write function for 8it buswith
 */
static void nand_write_byte(struct mtd_info *mtd, u_char byte)
{
	struct nand_chip *this = mtd->priv;
	writeb(byte, this->IO_ADDR_W);
}


/**
 * nand_read_byte16 - [DEFAULT] read one byte endianess aware from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 16bit buswith with
 * endianess conversion
 */
/* 
static u_char nand_read_byte16(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	return (u_char) cpu_to_le16(readw(this->IO_ADDR_R));
}
*/
/**
 * nand_write_byte16 - [DEFAULT] write one byte endianess aware to the chip
 * @mtd:	MTD device structure
 * @byte:	pointer to data byte to write
 *
 * Default write function for 16bit buswith with
 * endianess conversion
 */
/*
static void nand_write_byte16(struct mtd_info *mtd, u_char byte)
{
	struct nand_chip *this = mtd->priv;
	writew(le16_to_cpu((u16) byte), this->IO_ADDR_W);
}
*/

/**
 * nand_read_word - [DEFAULT] read one word from the chip
 * @mtd:	MTD device structure
 *
 * Default read function for 16bit buswith without
 * endianess conversion
 */
/* 
static u16 nand_read_word(struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	return readw(this->IO_ADDR_R);
}
*/

/**
 * nand_write_word - [DEFAULT] write one word to the chip
 * @mtd:	MTD device structure
 * @word:	data word to write
 *
 * Default write function for 16bit buswith without
 * endianess conversion
 */
/* 
static void nand_write_word(struct mtd_info *mtd, u16 word)
{
	struct nand_chip *this = mtd->priv;
	writew(word, this->IO_ADDR_W);
}
*/

/**
* Write 1 Byte to the pio
*/
void pio_nand_flash_write(unsigned char cValue)
{
  np_pio *pio=(np_pio*)(na_nand_flash_pio);

  outl(-1,&pio->np_piodirection);
  outl(cValue,&pio->np_piodata);
}

/**
* Read 1 Byte from the pio
*/
unsigned char pio_nand_flash_read(void)
{
  np_pio *pio=(np_pio*)(na_nand_flash_pio);
  unsigned char cHelp;

  outl(0,&pio->np_piodirection);
  cHelp=(unsigned char)inl(&pio->np_piodata);
  return(cHelp);
}

/**
* CLE...command latch enable, high active
*/
void pio_nand_flash_set_cle(void)
{
  unsigned char cHelp;

//  printk("set cle\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x09;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_cle(void)
{
  unsigned char cHelp;

//  printk("clr cle\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFE;
  pio_nand_flash_write(cHelp);
}

/**
* ALE...address latch enable, high active
*/
void pio_nand_flash_set_ale(void)
{
  unsigned char cHelp;

//  printk("set ale\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x0A;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_ale(void)
{
  unsigned char cHelp;

//  printk("clr ale\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFD;
  pio_nand_flash_write(cHelp);
}

/**
* CE...chip enable, low active
*/
void pio_nand_flash_set_nce(void)
{
  unsigned char cHelp;

//  printk("set nce\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xFB;
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_nce(void)
{
  unsigned char cHelp;

//  printk("clr nce\n");
  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x0C;                        //also set wp pin
  pio_nand_flash_write(cHelp);
}

/**
* WP...write protect, low active
*/
void pio_nand_flash_set_nwp(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp & 0xF7;
  pio_nand_flash_write(cHelp);
}

void pio_nand_flash_clr_nwp(void)
{
  unsigned char cHelp;

  cHelp=pio_nand_flash_read();
  cHelp=cHelp | 0x08;
  pio_nand_flash_write(cHelp);
}

/*
 *	hardware specific access to control-lines
 */
static void nand01gw3b2b_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl)
{
// static unsigned int OldCtrl = 0;

//  printk("dat=0x%04x (%3d), ctrl=0x%04x (%3d)\n", (unsigned int)dat, dat, ctrl, ctrl);

  if(ctrl & NAND_CTRL_CHANGE)
  {
 
    if(ctrl & NAND_NCE)
    {
//     if(!(OldCtrl & NAND_NCE))
     {    
      pio_nand_flash_set_nce();
     }     
    }
    else
    {
//     if(OldCtrl & NAND_NCE)
     {    
      pio_nand_flash_clr_nce();
     }     
    }
    
    if(ctrl & NAND_CLE)
    {
//     if(!(OldCtrl & NAND_CLE))
     {    
      pio_nand_flash_set_cle();
     }     
    }
    else
    {
//     if(OldCtrl & NAND_CLE)
     {    
      pio_nand_flash_clr_cle();
     }    
    }
    
    if(ctrl & NAND_ALE)
    {
//     if(!(OldCtrl & NAND_ALE))
     {
      pio_nand_flash_set_ale();
     }     
    }
    else
    {
//     if(OldCtrl & NAND_ALE)
     {    
      pio_nand_flash_clr_ale();
     }    
    }    
   
//   OldCtrl = ctrl;  
  }

	if (cmd == NAND_CMD_NONE)
		return;
    
  nand_write_byte(mtd, cmd);
  
/*
  //pio_nand_flash_clr_nwp();      //remove write protection
	switch(cmd) 
	{

	  case NAND_CTL_SETCLE:
      pio_nand_flash_set_cle();
		  break;
	  case NAND_CTL_CLRCLE:
      pio_nand_flash_clr_cle();
		  break;

	  case NAND_CTL_SETALE:
      pio_nand_flash_set_ale();
		  break;
	  case NAND_CTL_CLRALE:
      pio_nand_flash_clr_ale();
		  break;

	  case NAND_CTL_SETNCE:
      pio_nand_flash_set_nce();
		  break;
	  case NAND_CTL_CLRNCE:
      pio_nand_flash_clr_nce();
		  break;
	}
*/	
}

/*
 *	read device ready pin
 *  @return 0 = busy, 1 = ready
 */
#if 0
static int nand01gw3b2b_device_ready(struct mtd_info *mtd)
{
  return(0);
}
#endif

/*
 * Main initialization routine
 */
static int __init nand01gw3b2b_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	void __iomem *nandaddr;

  //allocate ioport for control signals
  request_mem_region((unsigned long)na_nand_flash_pio,sizeof(np_pio),"pio_nand_flash");

	nandaddr = ioremap(na_nand_flash, 1024);
	if (!nandaddr) {
		printk("Failed to ioremap nand flash.\n");
		return -ENOMEM;
	}

	/* Allocate memory for MTD device structure and private data */
	nand01gw3b2b_nand_mtd = kmalloc(sizeof(struct mtd_info) +
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!nand01gw3b2b_nand_mtd) {
		printk("Unable to allocate nand01gw3b2b NAND MTD device structure.\n");
		iounmap ((void *) nandaddr);
		return -ENOMEM;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *) (&nand01gw3b2b_nand_mtd[1]);

	/* Initialize structures */
	memset((char *) nand01gw3b2b_nand_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	nand01gw3b2b_nand_mtd->priv = this;
	nand01gw3b2b_nand_mtd->owner = THIS_MODULE;

	/* insert callbacks */
	this->IO_ADDR_R = nandaddr;
	this->IO_ADDR_W = nandaddr;
//	this->hwcontrol = nand01gw3b2b_hwcontrol;
	this->cmd_ctrl = nand01gw3b2b_hwcontrol;
	this->dev_ready = NULL;	/* unknown whether that was correct or not so we will just do it like this */
	/* 15 us command delay time */
	this->chip_delay = 50;
//	this->eccmode = NAND_ECC_SOFT;
	this->ecc.mode = NAND_ECC_SOFT;	
	//this->eccmode = NAND_ECC_NONE;

//	this->options = NAND_NO_AUTOINCR;
	
//	this->options |= NAND_BUSWIDTH_16;

	/* Scan to find existence of the device */
	if (nand_scan (nand01gw3b2b_nand_mtd, 1)) {
		printk(KERN_NOTICE "No NAND device - returning -ENXIO\n");
		kfree (nand01gw3b2b_nand_mtd);
		iounmap ((void *) nandaddr);
		return -ENXIO;
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(nand01gw3b2b_nand_mtd, partition_info, NUM_PARTITIONS);

	/* Return happy */
	return 0;
}
module_init(nand01gw3b2b_init);

/*
 * Clean up routine
 */
static void __exit nand01gw3b2b_cleanup (void)
{
	struct nand_chip *this = (struct nand_chip *) &nand01gw3b2b_nand_mtd[1];

	/* Release resources, unregister device */
	nand_release (nand01gw3b2b_nand_mtd);

	/* Release io resource */
	iounmap ((void *) this->IO_ADDR_W);

	/* Free the MTD device structure */
	kfree (nand01gw3b2b_nand_mtd);
  release_mem_region((unsigned long)na_nand_flash_pio,sizeof(np_pio));
}
module_exit(nand01gw3b2b_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Manfred Schmid <manfred at wolfvision dot com>");
MODULE_DESCRIPTION("NAND flash driver for fb4");
